/*
 * linux/drivers/ide/arm/ide-ep93xx.c
 * Copyright (C) 2007 Atmark Techno, Inc
 *
 * Adapted from mpc8xx.c, which is:
 *  Copyright (C) 2000, 2001 Wolfgang Denk, wd@denx.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 */

#include <linux/ide.h>
#include <linux/delay.h>
#include <linux/ioport.h>

#include <asm/io.h>
#include <asm/arch/irq.h>

/*
 * CIS Tupel codes
 */
#define CISTPL_NULL		0x00
#define CISTPL_DEVICE		0x01
#define CISTPL_LONGLINK_CB	0x02
#define CISTPL_INDIRECT		0x03
#define CISTPL_CONFIG_CB	0x04
#define CISTPL_CFTABLE_ENTRY_CB 0x05
#define CISTPL_LONGLINK_MFC	0x06
#define CISTPL_BAR		0x07
#define CISTPL_PWR_MGMNT	0x08
#define CISTPL_EXTDEVICE	0x09
#define CISTPL_CHECKSUM		0x10
#define CISTPL_LONGLINK_A	0x11
#define CISTPL_LONGLINK_C	0x12
#define CISTPL_LINKTARGET	0x13
#define CISTPL_NO_LINK		0x14
#define CISTPL_VERS_1		0x15
#define CISTPL_ALTSTR		0x16
#define CISTPL_DEVICE_A		0x17
#define CISTPL_JEDEC_C		0x18
#define CISTPL_JEDEC_A		0x19
#define CISTPL_CONFIG		0x1a
#define CISTPL_CFTABLE_ENTRY	0x1b
#define CISTPL_DEVICE_OC	0x1c
#define CISTPL_DEVICE_OA	0x1d
#define CISTPL_DEVICE_GEO	0x1e
#define CISTPL_DEVICE_GEO_A	0x1f
#define CISTPL_MANFID		0x20
#define CISTPL_FUNCID		0x21
#define CISTPL_FUNCE		0x22
#define CISTPL_SWIL		0x23
#define CISTPL_END		0xff

/*
 * CIS Function ID codes
 */
#define CISTPL_FUNCID_MULTI	0x00
#define CISTPL_FUNCID_MEMORY	0x01
#define CISTPL_FUNCID_SERIAL	0x02
#define CISTPL_FUNCID_PARALLEL	0x03
#define CISTPL_FUNCID_FIXED	0x04
#define CISTPL_FUNCID_VIDEO	0x05
#define CISTPL_FUNCID_NETWORK	0x06
#define CISTPL_FUNCID_AIMS	0x07
#define CISTPL_FUNCID_SCSI	0x08

/*
 * Fixed Disk FUNCE codes
 */
#define CISTPL_IDE_INTERFACE	0x01

#define CISTPL_FUNCE_IDE_IFACE	0x01
#define CISTPL_FUNCE_IDE_MASTER	0x02
#define CISTPL_FUNCE_IDE_SLAVE	0x03

/* First feature byte */
#define CISTPL_IDE_SILICON	0x04
#define CISTPL_IDE_UNIQUE	0x08
#define CISTPL_IDE_DUAL		0x10

/* Second feature byte */
#define CISTPL_IDE_HAS_SLEEP	0x01
#define CISTPL_IDE_HAS_STANDBY	0x02
#define CISTPL_IDE_HAS_IDLE	0x04
#define CISTPL_IDE_LOW_POWER	0x08
#define CISTPL_IDE_REG_INHIBIT	0x10
#define CISTPL_IDE_HAS_INDEX	0x20
#define CISTPL_IDE_IOIS16	0x40

static void ep93xx_pcmcia_print_funcid (int func)
{
	switch (func) {
	case CISTPL_FUNCID_MULTI:
		printk (" Multi-Function");
		break;
	case CISTPL_FUNCID_MEMORY:
		printk (" Memory");
		break;
	case CISTPL_FUNCID_SERIAL:
		printk (" Serial Port");
		break;
	case CISTPL_FUNCID_PARALLEL:
		printk (" Parallel Port");
		break;
	case CISTPL_FUNCID_FIXED:
		printk (" Fixed Disk");
		break;
	case CISTPL_FUNCID_VIDEO:
		printk (" Video Adapter");
		break;
	case CISTPL_FUNCID_NETWORK:
		printk (" Network Adapter");
		break;
	case CISTPL_FUNCID_AIMS:
		printk (" AIMS Card");
		break;
	case CISTPL_FUNCID_SCSI:
		printk (" SCSI Adapter");
		break;
	default:
		printk (" Unknown");
		break;
	}
	printk (" Card\n");
}

static void ep93xx_pcmcia_print_fixed (volatile u8 *p)
{
	if (p == NULL)
		return;

	switch (*p) {
	case CISTPL_FUNCE_IDE_IFACE:
	    {   u8 iface = *(p+2);

		printk ((iface == CISTPL_IDE_INTERFACE) ? " IDE" : " unknown");
		printk (" interface ");
		break;
	    }
	case CISTPL_FUNCE_IDE_MASTER:
	case CISTPL_FUNCE_IDE_SLAVE:
	    {   u8 f1 = *(p+2);
		u8 f2 = *(p+4);

		printk ((f1 & CISTPL_IDE_SILICON) ? " [silicon]" : " [rotating]");

		if (f1 & CISTPL_IDE_UNIQUE)
			printk (" [unique]");

		printk ((f1 & CISTPL_IDE_DUAL) ? " [dual]" : " [single]");

		if (f2 & CISTPL_IDE_HAS_SLEEP)
			printk (" [sleep]");

		if (f2 & CISTPL_IDE_HAS_STANDBY)
			printk (" [standby]");

		if (f2 & CISTPL_IDE_HAS_IDLE)
			printk (" [idle]");

		if (f2 & CISTPL_IDE_LOW_POWER)
			printk (" [low power]");

		if (f2 & CISTPL_IDE_REG_INHIBIT)
			printk (" [reg inhibit]");

		if (f2 & CISTPL_IDE_HAS_INDEX)
			printk (" [index]");

		if (f2 & CISTPL_IDE_IOIS16)
			printk (" [IOis16]");

		break;
	    }
	}
	printk ("\n");
}

#define	MAX_TUPEL_SZ	512
#define MAX_FEATURES	4

static int ep93xx_pcmcia_check_ide_device (unsigned long base)
{
	volatile u8 *ident = NULL;
	volatile u8 *feature_p[MAX_FEATURES];
	volatile u8 *p, *start;
	int n_features = 0;
	u8 func_id = ~0;
	u8 code, len;
	unsigned short config_base = 0;
	int found = 0;
	int i;

#ifdef DEBUG
	printk ("PCMCIA MEM: %08lX\n", base);
#endif
	start = p = (volatile u8 *) base;

	while ((p - start) < MAX_TUPEL_SZ) {

		code = *p; p += 2;

		if (code == 0xFF) { /* End of chain */
			break;
		}

		len = *p; p += 2;
#ifdef	DEBUG_PCMCIA
		{ volatile u8 *q = p;
			printk ("\nTuple code %02x  length %d\n\tData:",
				code, len);

			for (i = 0; i < len; ++i) {
				printk (" %02x", *q);
				q+= 2;
			}
		}
#endif	/* DEBUG_PCMCIA */
		switch (code) {
		case CISTPL_VERS_1:
			ident = p + 4;
			break;
		case CISTPL_FUNCID:
			func_id = *p;
			break;
		case CISTPL_FUNCE:
			if (n_features < MAX_FEATURES)
				feature_p[n_features++] = p;
			break;
		case CISTPL_CONFIG:
			config_base = (*(p+6) << 8) + (*(p+4));
		default:
			break;
		}
		p += 2 * len;
	}

	if (func_id != ((u8)~0)) {
		ep93xx_pcmcia_print_funcid (func_id);

		if (func_id == CISTPL_FUNCID_FIXED)
			found = 1;
		else
			return (1);	/* no disk drive */
	}

	for (i=0; i<n_features; ++i) {
		ep93xx_pcmcia_print_fixed (feature_p[i]);
	}

	if (!found) {
		printk ("unknown card type\n");
		return (1);
	}

	/* set level mode irq and I/O mapped device in config reg*/
	*((u8 *)(base + config_base + 6)) = 0x0;
	*((u8 *)(base + config_base + 0)) = 0x41;
	*((u8 *)(base + config_base + 2)) = 0x0;
	*((u8 *)(base + config_base + 4)) = 0x0;

	return (0);
}

#define EP93XX_PHYS_ADDR_PCMCIAMEM 	(PCMCIA_BASE_PHYS + 0xC000000)
#define EP93XX_SIZE_PCMCIAMEM		0x04000000

#define EP93XX_PHYS_ADDR_PCMCIAATTR	(PCMCIA_BASE_PHYS + 0x8000000)
#define EP93XX_SIZE_PCMCIAATTR		0x04000000

#define EP93XX_PHYS_ADDR_PCMCIAIO	(PCMCIA_BASE_PHYS)
#define EP93XX_SIZE_PCMCIAIO		0x04000000

#define EP93XX_PCMCIA_WP		0x01
#define EP93XX_PCMCIA_CD1		0x02
#define EP93XX_PCMCIA_CD2		0x04
#define EP93XX_PCMCIA_BVD1		0x08
#define EP93XX_PCMCIA_BVD2		0x10
#define EP93XX_PCMCIA_VS1		0x20
#define EP93XX_PCMCIA_RDY		0x40
#define EP93XX_PCMCIA_VS2		0x80

#define EP93XX_PCMCIA_INT_WP		IRQ_GPIO0	/* 19 */
#define EP93XX_PCMCIA_INT_CD1		IRQ_GPIO1	/* 20 */
#define EP93XX_PCMCIA_INT_CD2		IRQ_GPIO2	/* 21 */
#define EP93XX_PCMCIA_INT_BVD1		IRQ_GPIO3	/* 22 */
#define EP93XX_PCMCIA_INT_BVD2		IRQ_GPIO4	/* 47 */
#define EP93XX_PCMCIA_INT_VS1		IRQ_GPIO5	/* 48 */
#define EP93XX_PCMCIA_INT_RDY		IRQ_GPIO6	/* 49 */
#define EP93XX_PCMCIA_INT_VS2		IRQ_GPIO7	/* 50 */

static void ep93xx_pcmcia_ide_interrupt_ack (void *dev)
{
	unsigned long ulTemp;

	outl( EP93XX_PCMCIA_RDY, GPIO_FEOI );
	ulTemp = inl( GPIO_FEOI );
}

static void *pcmcia_base = NULL;

void ep93xx_pcmcia_ide_init (void)
{
	hw_regs_t hw;
	unsigned long ulPFDR;
	unsigned long ulTemp;
	void *pcmcia_attr = NULL;

	if (!pcmcia_base) {
#ifdef CONFIG_ARCH_ARMADILLO9
		outl( inl(GPIO_PADR) | 0x8, GPIO_PADR );	//Power off
		mdelay(300);
#endif
		pcmcia_base = ioremap( EP93XX_PHYS_ADDR_PCMCIAIO, 0x10000 );
		ulTemp = inl(GPIO_FINTEN);
		outl( ulTemp & ~(EP93XX_PCMCIA_RDY | EP93XX_PCMCIA_CD1 |
				 EP93XX_PCMCIA_CD2 | EP93XX_PCMCIA_BVD2 |
				 EP93XX_PCMCIA_BVD1),
		      GPIO_FINTEN);
		ulTemp = inl(GPIO_FINTEN);
		ulTemp = inl(GPIO_PFDDR);
		outl( ulTemp & ~(EP93XX_PCMCIA_RDY | EP93XX_PCMCIA_CD1 |
				 EP93XX_PCMCIA_CD2 | EP93XX_PCMCIA_BVD2 |
				 EP93XX_PCMCIA_BVD1),
		      GPIO_PFDDR);
		ulTemp = inl(GPIO_PFDDR);
		ulTemp = inl(GPIO_FDB);
		outl( (ulTemp & ~(EP93XX_PCMCIA_RDY | EP93XX_PCMCIA_BVD2 |
				  EP93XX_PCMCIA_BVD1)) |
		      EP93XX_PCMCIA_CD1 | EP93XX_PCMCIA_CD2,
		      GPIO_FDB );
		ulTemp = inl(GPIO_FINTTYPE1);
		outl( (ulTemp & ~EP93XX_PCMCIA_RDY) |
		      EP93XX_PCMCIA_CD1 | EP93XX_PCMCIA_CD2 |
		      EP93XX_PCMCIA_BVD2 | EP93XX_PCMCIA_BVD1,
		      GPIO_FINTTYPE1 );
		ulTemp = inl(GPIO_FINTTYPE2);
		outl( ulTemp & ~(EP93XX_PCMCIA_RDY | EP93XX_PCMCIA_CD1 |
				 EP93XX_PCMCIA_CD2 | EP93XX_PCMCIA_BVD2 |
				 EP93XX_PCMCIA_BVD1),
		      GPIO_FINTTYPE2 );
		ulTemp = inl(GPIO_FINTTYPE2);
		outl( EP93XX_PCMCIA_RDY | EP93XX_PCMCIA_CD1 |
		      EP93XX_PCMCIA_CD2 | EP93XX_PCMCIA_BVD2 |
		      EP93XX_PCMCIA_BVD1,
		      GPIO_FEOI );
		ulTemp = inl(GPIO_FEOI);
		outl( 0xa0405, SMC_PCIO );
		ulTemp = inl(SMC_PCIO);
		outl( 0x110403, SMC_PCAttribute );
		ulTemp = inl(SMC_PCAttribute);
		outl( 0x110403, SMC_PCCommon );
		ulTemp = inl(SMC_PCCommon);
		ulTemp = (inl(SMC_PCMCIACtrl) | PCCONT_WEN | PCCONT_PC1EN) &
		  ~PCCONT_PC1RST;
		outl( ulTemp, SMC_PCMCIACtrl );
		ulTemp = inl(SMC_PCMCIACtrl);
		ulTemp = inl(GPIO_FINTEN);
		outl( (ulTemp & ~(EP93XX_PCMCIA_CD1 | EP93XX_PCMCIA_CD2 |
				  EP93XX_PCMCIA_BVD2 | EP93XX_PCMCIA_BVD1)) |
		      EP93XX_PCMCIA_RDY,
		      GPIO_FINTEN );
		ulTemp = inl(GPIO_FINTEN);
	}

	ulPFDR = inl(GPIO_PFDR);

#ifdef DEBUG
	printk ("PFDR = 0x%08X  mask = 0x%X\n",
		ulPFDR,
		ulPFDR & (EP93XX_PCMCIA_CD1 | EP93XX_PCMCIA_CD2) );
#endif /* DEBUG */

	if (ulPFDR & (EP93XX_PCMCIA_CD1 | EP93XX_PCMCIA_CD2)) {
		printk ("No card in slot: PFDR=%08x\n", (unsigned int)ulPFDR);
		return;		/* No card in slot */
	}

#ifdef CONFIG_ARCH_ARMADILLO9
	outl( inl(GPIO_PADR) & ~0x8, GPIO_PADR );	//Power on
	mdelay(300);
#endif

	ulTemp = inl(SMC_PCMCIACtrl) |
	  PCCONT_WEN | PCCONT_PC1EN | PCCONT_PC1RST;
	outl( ulTemp, SMC_PCMCIACtrl );
	ulTemp = (inl(SMC_PCMCIACtrl) | PCCONT_WEN | PCCONT_PC1EN) &
	  ~PCCONT_PC1RST;
	udelay(10);
	outl( ulTemp, SMC_PCMCIACtrl );
	ulTemp = inl(SMC_PCMCIACtrl);
	mdelay(500);

	pcmcia_attr = ioremap( EP93XX_PHYS_ADDR_PCMCIAATTR, 0x00001000 );
	if (ep93xx_pcmcia_check_ide_device ((unsigned long)pcmcia_attr)) {
#ifdef CONFIG_ARCH_ARMADILLO9
		outl( inl(GPIO_PADR) | 0x8, GPIO_PADR );	//Power off
		mdelay(300);
#endif
		iounmap( pcmcia_attr );
		return;
	}
	iounmap( pcmcia_attr );

	release_region((int)pcmcia_base, 0x10);
	outb(0x02, ((unsigned long)pcmcia_base) + 0xe);
	ide_init_hwif_ports
	  (&hw, (int)pcmcia_base, ((int)pcmcia_base) + 0xe, NULL);
	hw.ack_intr = (ide_ack_intr_t *)ep93xx_pcmcia_ide_interrupt_ack;
	hw.irq = EP93XX_PCMCIA_INT_RDY;
	hw.chipset = ide_pci;
	ide_register_hw(&hw, NULL);
}

